use std::time::Duration;

use sfml::system::Vector2;
use sfml::window::Event;
use sfml::graphics::View;

use engine::character::Character;
use engine::map::Map;

/// Contains everything needed to play.
pub struct Scene {

    /// The characters contained in the scene.
    characters: Vec<Character>,

    /// The map of the scene.
    map: Map,

}

impl Scene {

    /// Creates an empty scene.
    pub fn new() -> Scene {
        Scene {
            characters: vec![],
            map: Map::from_str(include_str!("../../../assets/levels/level2.txt")).unwrap(),
        }
    }

    /// Adds a character to the scene.
    pub fn add(&mut self, character: Character) {
        let mut character = character;
        character.position.x = self.map.entrance().1 as f32 * 16.0;
        character.position.y = self.map.entrance().0 as f32 * 16.0;

        self.characters.push(character);
    }

    /// Returns the controlable.
    fn controlable(&self) -> Option<&Character> {
        for character in &self.characters {
            if character.controls().is_some() {
                return Some(&character)
            }
        }
        None
    }

    /// Returns the right view.
    pub fn view(&self) -> Option<View> {
        let view = self.controlable()?.view();
        let mut center = view.center();
        let size = view.size();

        // Clamp center so that the view doesn't show things outside the level.
        if center.x - size.x / 2.0 < 0.0 {
            center.x = size.x / 2.0;
        }

        if center.y - size.y / 2.0 < 0.0 {
            center.y = size.y / 2.0;
        }

        let right_limit = self.map.cols() as f32 * 16.0;
        let bottom_limit = self.map.rows() as f32 * 16.0;

        if center.x + size.x / 2.0 > right_limit {
            center.x = right_limit - size.x / 2.0;
        }

        if center.y + size.y / 2.0 > bottom_limit {
            center.y = bottom_limit - size.y / 2.0;
        }

        Some(View::new(center, size))
    }

    /// Updates the whole scene.
    pub fn update(&mut self, duration: &Duration) {

        for c in &mut self.characters {
            let old = c.bbox();

            // Compute the offset between position and bbox
            let offset = Vector2::new(old.left, old.top) - c.position;

            c.update(duration);

            if let Some((axis, position)) = self.map.collides_bbox(old, c.bbox()) {
                c.position = position - offset;
                if axis.is_x() {
                    c.speed.x = 0.0;
                }
                if axis.is_y() {
                    c.speed.y = 0.0;
                    c.ground_collision();
                }
            } else {
                c.fall_off();
            }
        }
    }

    /// Transfers an event to the elements contained in the scene that should receive events.
    pub fn manage_event(&mut self, event: &Event) {
        for c in &mut self.characters {
            c.manage_event(event);
        }
    }

    /// Returns a reference to the characters of the scene.
    pub fn characters(&self) -> &Vec<Character> {
        &self.characters
    }

    /// Returns a reference to the map.
    pub fn map(&self) -> &Map {
        &self.map
    }

}

/// Trait that needs to be implemented for everything that can be updatable.
pub trait Updatable {
    /// Updates the thing depending on the duration since last frame.
    fn update(&mut self, duration: &Duration);

    /// Called when an event arrives.
    fn manage_event(&mut self, event: &Event);
}
